#ifndef GLOBALS_H_
#define GLOBALS_H_

#include "project.def"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef CAN_SENDGLOBALS
# define CAN_SENDGLOBALS TRUE /* default behavior for opening the globals socket */
#endif
#ifndef GLOBALGRACE
# define GLOBALGRACE  3      /* retries for missing heartbeat */
#endif

/**
 * \internal
 * \brief Structure used to managed global data.
 */
#define KEYWORD_NAME_SHORTSIZE 16
#define KEYWORD_NAME_SIZE 32
typedef struct         /* net */
{
  char  name[KEYWORD_NAME_SHORTSIZE];      /* global element name  (16 bytes) */
  short size;          /* number of elements gcast under "name" (2 bytes) */
  short format;        /* TINE data format  (2 bytes) */
  SINT32 time;         /* timestamp of gcast data (4 bytes) */
} GlobalDataHdr;
#define GLOBALDATA_HEADER_SIZE (KEYWORD_NAME_SHORTSIZE + 2 + 2 + 4)

typedef struct
{
  char context[CONTEXT_NAME_SIZE]; /* global context */
  char keyword[KEYWORD_NAME_SIZE]; /* global keyword */
  short size;                /* number of elements registered in global array */
  short format;              /* globals data format */
  time_t time;               /* timestamp of globals data */
  struct timeval tv;         /* timestamp of globals data */
  BYTE *data;                /* data pointer */
  BYTE *refdata;             /* for comparison */
  int  minPeriod,            /* minimum period between multicasts */
       maxPeriod;            /* maximum period between multicasts */
  double tolerance;          /* allowed deviation of data (data units) */
  UINT32 mcastGrp;           /* the multicast group ip addr */
  UINT32 callbackID;         /* private callback ID */
  void (*fcn)(int,int);      /* call back function */
  short isPending;           /* pending flag */
  short isGrouped;           /* grouped flag */
  short disabled;            /* disabled flag */
  short hasUserCallbackId;   /* flag : TRUE if set by caller, otherwise false */
  short lastsize;            /* number of elements in last incoming/outgoing packet */
} GlobalListStruct;

#ifndef GLOBALSRV
# define GLOBALSRV "GLOBALS"
#endif
#ifndef MAXGLOBALS
# define MAXGLOBALS 50
#endif
#ifndef GLOBALHEARTBEAT
# define GLOBALHEARTBEAT 60
#endif

extern int nglobals, ngcasts;
extern int GlobalHeartbeat;
extern int MaxNumGlobals;

extern GlobalListStruct **gcastTbl;   /* the server's send list */
extern GlobalListStruct **glbTbl; /* the client's recv list */

/**
 * \internal
 */
int parseGlobalsKeyword(char *key,char **context,char **keyword);
/**
 * \internal
 */
int getGlobalTableIndex(char *context,char *keyword);

/* sendNetGlobal() */
/**
 * \brief registers and sends the accompanying keyword and data as a network global.
 *
 * The keyword supplied is appended to the globals table if not already registered and sent
 * immediately as a netork global per multicast if multicasting is enabled and/or per broadcast
 * if a broadcast nets table has been registered (via the inclusion of a ipbcast.csv or ipxbcast.csv
 * file in the startup database set).  Multicasts/Broadcasts will then be scheduled according the the minPeriod,
 * maxPeriod, and tolerance specified, in which case the din data object should contain a reference 
 * to a data object which will be regularly updated by the server code.  Alternatively, the 
 * call can specify '-1' as minPeriod and maxPeriod, in which case the network global will only be sent
 * upon a call to sendNetGlobal().
 * If a callback function is supplied, is it called following the successful transmission of the
 * network global.
 * 
 * Returns 0 upon successful globals transmission, otherwise a TINE error code.
 *
 * @param keyword is the globals KEYWORD which identifies the data set being sent.
 * @param din is a pointer to the input data set, that is, the data set to be sent from the server 
 * If din is a NULL pointer or contains a NULL data set, sendNetGlobal() returns an error.
 * @param callback is the address of the user-supplied callback routine to be called when data are sent 
 * from the server.
 * The prototype of this routine must be as shown, where the first argument (id) gives the id of the link 
 * which is calling the callback, and the second argument (cc) gives the completion code.  For successful data 
 * transmission, cc will always be 0, otherwise it can be interpreted with GetLastLinkError().
 * @param minPeriod is the minimum period in milliseconds between scheduled multi/broadcasts
 * @param maxPeriod is the maximum period in milliseconds between scheduled multi/broadcasts.  If the data are
 * still within tolerance at the end of this period, the data set will nonetheless be sent as a network
 * global.
 * @param tolerance is the allowed range between the current and previous data sets.  If exceeded, the
 * current data set is then sent as a network global.
 *  
 */
TINE_EXPORT int sendNetGlobal(char *keyword,DTYPE *din,void (*callback)(int,int),
                  int minPeriod,int maxPeriod,double tolerance);

/**
 * \brief retrieves the link ID for a globals link
 *
 * The ID supplied as argument is the requested callback ID from the original link
 * registration.  This link ID is needed as argument to CloseLink().
 * 
 *
 * @param cbId is the callback ID passed as an argument to recvNetGlobalEx().
 *  
 * \return the link table ID for the globals link or a negative TINE error code.
 */
TINE_EXPORT int GetGlobalLinkIdFromCallbackId(int cbId);
#define getGlobalLinkIdFromCallbackId(a1) GetGlobalLinkIdFromCallbackId(a1)

/**
 * \brief retrieves the link ID for a globals link
 *
 * @param keyword is the globals keyword passed as an argument to recvNetGlobal().
 *  
 * \return the link table ID for the globals link or a negative TINE error code.
 */
TINE_EXPORT int GetGlobalLinkId(char *keyword);
#define getGlobalLinkId(a1) GetGlobalLinkId(a1)

/**
 * \internal
 */
TINE_EXPORT int ListenForMulticasts(void);

/**
 * \internal
 */
TINE_EXPORT int ListenForNetGlobals(void);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* GLOBALS_H_ */
