#ifndef PROLOG_H_
#define PROLOG_H_

#include "project.def"
#include "tinedefs.h"
#include "errors.h"
#include "swaplib.h"
#include "tinetype.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

/**
 * \brief Allocate memory for an array. Synonym for calloc() under normal circumstances.
 *
 * When some other form of memory managment is in place (e.g. in VxWorks in an EPICS context)
 * SystemCalloc can be re#defined in project.def to use it.
 *
 * \param nr Number of elements.
 * \param sz Size of an element in bytes.
 */
#ifndef SystemCalloc
# define SystemCalloc(nr,sz) calloc(nr,sz)
#endif

/**
 * \brief Free allocated memory. Synonym for free() under normal circumstances.
 *
 * When some other form of memory managment is in place (e.g. in VxWorks in an EPICS context)
 * SystemCalloc can be re#defined in project.def to use it.
 *
 * \param ptr pointer to memory area to be free()ed.
 */
#ifndef SystemFree
# define SystemFree(ptr) free(ptr)
#endif

#ifndef FEC_TIMEOUT
#define FEC_TIMEOUT      1   /* seconds */
#endif
#ifndef SERVER_TIMEOUT
#define SERVER_TIMEOUT   1   /* seconds */
#endif
#ifndef MAX_FDSETS
#define MAX_FDSETS      64
#endif
#ifndef PORTOFFSET
# define PORTOFFSET 0
#endif
#ifndef WATCHDOG
# define WATCHDOG 0
#endif

/* OS-specific definitions ... */
#if defined(MACOS)
#  define UNIX
#  define FREEBSD
#  define OS_TAG "MACOS"
#endif
#if defined(DOS)              /* DOS */
#  define OS_TAG "DOS"
#  define FS_DELIMITER '\\'
#  ifndef NumTCPChan
#    define NumTCPChan	  5
#  endif
#  define SERVER_POLLING_MODE
#  define millisleep(ms) delay(ms)
   struct timezone { int dummy; };
#elif defined(UNIX)              /* UNIX */
#  ifndef OS_TAG
#    define OS_TAG "UNIX"
#  endif
#  define FS_DELIMITER '/'
#  ifndef NumTCPChan
#    define NumTCPChan	  16
#  endif
#  ifndef _REENTRANT
#    define _REENTRANT
#  endif
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(char *)buf)
#  define soclose(i) close(i)
   int millisleep(int msecs);
#elif defined(VXWORKS)         /* VxWorks */
#  define OS_TAG "VXWORKS"
#  define FS_DELIMITER '/'
#  define SCAN_FOR_NAME_LISTS FALSE
#  ifndef NumTCPChan
#    define NumTCPChan	  8
#  endif
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(int)buf)
#  define soclose(i) close(i)
   int millisleep(int msecs);
#elif defined(VMS)        /* VMS */
#  define OS_TAG "VMS"
#  ifndef NumTCPChan
#    define NumTCPChan	  8
#  endif
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) 0
#  define unlink(nm) remove(nm)
#  define millisleep(ms) ;
#  define soclose(i) close(i)
#  define shutdown(s,t) close(s)
   struct timezone { int dummy; };
#elif defined(WIN32)         /* Win32 */
#  define OS_TAG "WIN32"
#  define FS_DELIMITER '\\'
#  ifndef NumTCPChan
#    define NumTCPChan	  16
#  endif
#  define soerrno GetErrno()
#  define SERVER_SELECT_MODE
#  define soioctl(fd,REQ,buf) ioctlsocket(fd,REQ,(u_long *)buf)
#  define millisleep(ms) Sleep(ms)
#  define soclose(i) closesocket(i)
   struct timezone { int dummy; };
#elif defined(NIOS)          /* Altera NIOS */
#  define OS_TAG "NIOS"
#  define FS_DELIMITER '/'
#  define FS_RDONLY
#  ifndef NumTCPChan
#    define NumTCPChan	  5
#  endif
#  define ForceStructAlignment -1
#  define SERVER_POLLING_MODE
#  define millisleep(ms) sleep(ms)
#  define soioctl(fd,REQ,buf) ioctl(fd,REQ,(u_long *)buf)
#  define soread(fd,buf,nbyte)  recv(fd,(char *)buf,nbyte,0)
#  define sowrite(fd,buf,nbyte) send(fd,(char *)buf,nbyte,0)
   struct timezone { int dummy; };
#else                       /* ????? */ 
#  ifndef OS_TAG
#    define OS_TAG "UNKNOWN"
#  endif
   struct timezone { int dummy; };
#endif

#ifndef SCAN_FOR_DEVICE_LISTS
# define SCAN_FOR_DEVICE_LISTS TRUE
#endif

#ifndef SCAN_FOR_NAME_LISTS
# define SCAN_FOR_NAME_LISTS TRUE
#endif

#ifndef ForceStructAlignment
# define ForceStructAlignment 0
#endif

#ifndef soerrno
#  define soerrno errno
#endif

#ifndef SERVER_POLLING_MODE
#  define SERVER_SELECT_MODE
#endif

#ifdef MULTITHREADED
# if defined(WIN32)
#   define WINTHREAD
# elif defined (UNIX)
#   define PTHREAD
# endif
# include "threader.h"
# define nDataOutBuffers 2
#else
# define nDataOutBuffers 1
#endif

/* 'us-or-them' definitions : */
#ifdef DOS   
#  include "dos.h"
#  define malloc_t void *
#  define SetTimeOut(TimeOut) { TimeOut = SERVER_TIMEOUT; }
#  define RDATE
#  define srvWorkArea     WorkArea
#  define srvWorkAreaSize WorkAreaSize
#  define clnWorkArea     WorkArea
#  define clnWorkAreaSize WorkAreaSize
#  define whoami(c) ;
#else
#  define SetTimeOut(TimeOut) { TimeOut = SERVER_TIMEOUT+size/5000; }
#  define soperror(st) perror(st)
#  ifndef soread
#    define soread(fd,buf,nbyte)  read(fd,(char *)buf,nbyte)
#  endif
#  ifndef sowrite
#    define sowrite(fd,buf,nbyte) write(fd,(char *)buf,nbyte)
#  endif
#  if !defined(SUNOS)
#    define malloc_t size_t
#  endif
#  if !defined(WIN32)
   char *strrev(char *s);
#  endif
#  if !defined(WIN32) && !defined(NIOS)
   int stricmp(char *s1, char *s2); /* from the DOS world */
   int strnicmp(char *s1, char *s2,int len); /* from the DOS world */
   char *strupr(char *s);
   char *strlwr(char *s);
#  endif
   extern BYTE *srvWorkArea;
   extern UINT32 srvWorkAreaSize;
   extern BYTE *clnWorkArea;
   extern UINT32 clnWorkAreaSize;
   char *whoami(char *user);
   void ClientDelay(int sec);  /* delay(0) never returns */
#endif                         /* DOS: definitions */

#ifndef VXWORKS
#  define inet_ntoa_b(n,b) strcpy(b,inet_ntoa(n))
#else
  struct hostent
  {
    char *h_name;
    char **h_aliases;
    int  h_addrtype;
    int  h_length;
    char *h_addr;
  };
  extern struct hostent globalHostent;
  extern char globalHostAddr[5];
#endif

#ifndef UNIX
  TINE_EXPORT struct timeval *gettimeofday(struct timeval *t,struct timezone *tz);
#endif

void closeIPClients(void);

#ifndef IPXNW
typedef struct  /* an important entry from nwipxspx.h */
{
   BYTE network[4];
   BYTE node[6];
   BYTE socket[2];
} IPXAddress;
#endif

#define IsMemberIPControlNets(addr) IsMemberIPNets(NrOfIPNets,IPNetList,addr)
#define IsMemberIPBcastNets(addr) IsMemberIPNets(NrOfIPBcastNets,IPBcastList,addr)

#if defined(IPXNW) || defined(IPXBSD)
#  define IsMemberIPXControlNets(addr) IsMemberIPXNets(NrOfIPXNets,IPXNetList,addr)
#  define IsMemberIPXBcastNets(addr) IsMemberIPXNets(NrOfIPXBcastNets,IPXBcastList,addr)
#  define sendGCast(buf,size)  sendIPgcast(buf,size); sendIPXgcast(buf,size)
#  define AcceptRequest(tv) { AcceptIP(tv); AcceptIPX(tv); }
#  define Cleanup(tv)       { CleanupIPX(tv); CleanupIP(tv); CleanupStrandedDatagrams();}
#  define ntoa(c,b)     strcpy(b,\
   (*(UINT16 *)(c)->IPXaddress.socket ? ipxnet_ntoa((c)->IPXaddress.node) :\
				     inet_ntoa((c)->IPaddress.sin_addr)   )\
   )
#  define tCall(i,buf,to) (TCPloaded ? tUDPCall(i,buf,to) : tIPXCall(i,buf,to))
#  define IsMemberControlNets(c) ( *(DWORD *)(c)->IPXaddress.network ?\
				   IsMemberIPXControlNets(&(c)->IPXaddress) :\
				   IsMemberIPControlNets(&(c)->IPaddress)    )
#  define IsMemberBcastNets(c)   ( *(DWORD *)(c)->IPXaddress.network ?\
				   IsMemberIPXBcastNets(&(c)->IPXaddress) :\
				   IsMemberIPBcastNets(&(c)->IPaddress)    )
   void closeIPXClients(void);
   int initIPXClient(int socket);
   int sendSPXData(WORD ID,int socket);
#  define ShutdownClients(i) {\
     closeIPXClients();\
     closeIPClients();\
   }
#  define CheckConsumerData(i) {\
     if (i == 0) {ReturnToServerCycle++; saveClientCycle = ReturnToClientCycle;}\
     doIPXConsumer();\
     doUDPConsumer();\
     if (i == 0) {ReturnToServerCycle--; ReturnToClientCycle = saveClientCycle;}\
   }
#else
#  define sendGCast(buf,size) sendIPgcast(buf,size)
#  define AcceptRequest(tv) AcceptIP(tv)
#  define Cleanup(tv)       { CleanupIP(tv); CleanupStrandedDatagrams(); }
#  define ntoa(c,b)         inet_ntoa_b((c)->IPaddress.sin_addr,b)
#  define tCall(i,buf,to)   tUDPCall(i,buf,to)
#  define IsMemberControlNets(c) IsMemberIPControlNets(&(c)->IPaddress)
#  define IsMemberBcastNets(c) IsMemberIPBcastNets(&(c)->IPaddress)
#  define sendSPXData(i,j) ipx_not_supported
#  define ShutdownClients(i) {\
	   closeIPClients();\
   }
#  define CheckConsumerData(i) {\
     if (i == 0) {ReturnToServerCycle++; saveClientCycle = ReturnToClientCycle;}\
     doUDPConsumer();\
     if (i == 0) {ReturnToServerCycle--; ReturnToClientCycle = saveClientCycle;}\
   }
#endif         /* ! IPXNW || IPXBSD */

extern TINE_EXPORT int tineDebug;
extern int NGdebug;
int serverCycle(void);
int initTineResources(void);

#if defined(INTEL) || defined(VMS) || defined(NONALIGNED)
#  define PeekWord(w)    (w)
#  define PeekDWord(dw)  (dw)
#  define PrepWord(d,s)  { d = s; }
#  define PrepDWord(d,s) { d = s; }
#else
#  define PeekWord(w)    (short)ISWAP(w)
#  define PeekDWord(dw)  LSWAP(dw)
#  define PrepWord(d,s)  { d = (short)ISWAP(s); }
#  define PrepDWord(d,s) { d = LSWAP(s); }
#endif

#ifdef DOS  /* user-defined structure swapping */
# define AssignStructDataPtr(dptr,sptr,dir) { dptr = sptr;\
  switch (dir) {\
    case RTOL: i = (int)con->EqmSizeIn;\
      if ((s=getStructFromTag(con->strTagIn)) == NULL) break;\
      if ((i=i/s->rawlen) > s->num) return out_of_range;\
      if (con->EqmSizeIn%s->rawlen) return invalid_structure_tag;\
      break;\
    case LTOR: i = (int)con->EqmSizeOut;\
      if ((s=getStructFromTag(con->strTagOut)) == NULL) break;\
      if ((i=i/s->rawlen) > s->num) return out_of_range;\
      if (con->EqmSizeOut%s->rawlen) return invalid_structure_tag;\
      i = dout.dArrayLength; dout.dArrayLength *= s->rawlen;\
      break;\
    case NONE: i = (int)con->EqmSizeOut;\
      if ((s=getStructFromTag(con->strTagOut)) == NULL) break;\
      if ((i=i/s->rawlen) > s->num) return out_of_range;\
      if (con->EqmSizeOut%s->rawlen) return invalid_structure_tag;\
      break;\
  }\
  }
# define PrepOutgoingCLNQS(buf,c) memcpy(buf,c,sizeof(CLNQS));
#else
# define AssignStructDataPtr(dptr,sptr,dir) { dptr = sptr;\
  switch (dir) {\
    case RTOL: i = con->EqmSizeIn;\
      if (i == 0 || (s=getStructFromTag(con->strTagIn)) == NULL) break;\
      if ((i=i/s->rawlen) > s->num) return out_of_range;\
      if (con->EqmSizeIn%s->rawlen) return invalid_structure_tag;\
      StructSwapEx(con->strTagIn,(BYTE *)(sptr),s->sdata,RTOL,i,s->size);\
      dptr = s->sdata;\
      break;\
    case LTOR: i = con->EqmSizeOut;\
      if (i == 0 || (s=getStructFromTag(con->strTagOut)) == NULL) break;\
      if ((i=i/s->rawlen) > s->num) return out_of_range;\
      if (con->EqmSizeOut%s->rawlen) return invalid_structure_tag;\
      i = dout.dArrayLength; dout.dArrayLength *= s->rawlen;\
      StructSwapEx(con->strTagOut,(BYTE *)(dptr),s->sdata,LTOR,i,s->size);\
      break;\
    case NONE: i = con->EqmSizeOut;\
      if (i == 0 || (s=getStructFromTag(con->strTagOut)) == NULL) break;\
      if ((i=i/s->rawlen) > s->num) return out_of_range;\
      if (con->EqmSizeOut%s->rawlen) return invalid_structure_tag;\
      memset(s->sdata,0,i*s->size);\
      dptr = s->sdata;\
      break;\
  }\
  }

# define PrepOutgoingCLNQS(buf,c) {\
   _StructSwap(ClientQueryStructStruct,(unsigned char *)buf,(unsigned char *)c,LTOR);\
  }
#endif /* DOS */

#ifdef INTEL  /* Data Preparation MACROS: */

# define CopySubData(pD,con,siz,dir) {\
  if (dir == LTOR) memcpy(pD,(con)->DataOut,(size_t)siz);\
  else memcpy((con)->DataIn,pD,(size_t)siz);\
}

# define PrepSubData(pD,sz,f,dir) ;

# define PrepIncomingHeader(h,dat) { h = (HeaderStruct *)dat; }
# define PrepOutgoingAMS(amsstruct,buf,n) {if ((void *)buf != (void *)amsstruct) memcpy(buf,amsstruct,n*AMS_SIZE); }
# define PrepOutgoingAQS(a,buf) {if ((void *)buf != (void *)a) memcpy(buf,a,ACTIVITYQUERYSTRUCT_SIZE); }
# define PrepOutgoingPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PROPERTYQUERYSTRUCT_SIZE); }
# define PrepOutgoingPQSX(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PROPERTYQUERYSTRUCTEX_SIZE); }
# define PrepOutgoingXPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*XPROPERTYQUERYSTRUCT_SIZE); }
# define PrepOutgoingPRPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(buf,p,n*PRPQUERYSTRUCT_SIZE); }
# define PrepOutgoingCONQS(c) ;
# define PrepIncomingPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PROPERTYQUERYSTRUCT_SIZE); }
# define PrepIncomingPQSX(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PROPERTYQUERYSTRUCTEX_SIZE); }
# define PrepIncomingXPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*XPROPERTYQUERYSTRUCT_SIZE); }
# define PrepIncomingPRPPQS(p,buf,n) {if ((void *)buf != (void *)p) memcpy(p,buf,n*PRPQUERYSTRUCT_SIZE); }
# define PrepIncomingHRS(rb,buf) memcpy(rb,buf,HRS_SIZE);

# define AssignClnStructDataPtr(dst,src,dir) { dst = (BYTE *)src; }
# define cpyClnStructData(dst,src,siz) memcpy(dst,src,siz)
# define AssignOutgoingHeader(h,buf) {\
    if (!ForceStructAlignment) h = (HeaderStruct *)buf;\
    memset(h,0,sizeof(HeaderStruct));\
  }
# define AssignOutgoingBrdcstData(h,buf) h = (GlobalDataHdr *)buf; memset(h,0,sizeof(GlobalDataHdr));
# define PrepIncomingData(dst,src,dsiz,rsiz,fmt) {\
    if (fmt == CF_STRUCT && c != NULL && (s=getStructFromTag(c->strTagOut))) {\
      StructSwapEx(c->strTagOut,(BYTE *)src,(BYTE *)dst,RTOL,(int)rsiz/s->rawlen,s->size);\
    } else {\
      memcpy(dst,src,MIN(dsiz,rsiz));\
      if ((dsiz) > (rsiz)) memset(&((char *)(dst))[rsiz],0,dsiz-rsiz);\
    }\
    if (c != NULL && c->field[0] != 0) mapBitfieldToTarget(c->context,c->expName,c->strTagOut,c->field,fmt,dst);\
  }
# define PrepOutgoingHeader(h,dat) { if (ForceStructAlignment) memcpy(dat,h,HEADERSTRUCT_SIZE); }
# define PrepOutgoingGCastData(h,dat) ;
# define PrepIncomingGCastData(h,dat) h = (GlobalDataHdr *)dat;
# define PrepOutgoingData(d,s,siz,fmt) memcpy(d,s,siz)

#else     /* non-INTEL world must SWAP ! */

# define PrepSubData(pD,sz,f,dir) {\
    SwapBytes((char *)pD,sz,LFMT(f),dir);\
}
# define PrepIncomingHeader(h,dat) {\
    _StructSwap(HeaderStructStruct,(unsigned char *)dat,(unsigned char *)h,RTOL);\
}
# define PrepOutgoingAMS(amsstruct,buf,n)  {\
    _StructSwapEx(AMSStructStruct,(unsigned char *)buf,(unsigned char *)amsstruct,LTOR,n,sizeof(AMS));\
}
# define PrepOutgoingAQS(a,buf) {\
    _StructSwap(ActivityQueryStructStruct,(unsigned char *)buf,(unsigned char *)a,LTOR);\
}
# define PrepOutgoingPQS(p,buf,n) {\
    _StructSwapEx(PropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PropertyQueryStruct));\
}
# define PrepOutgoingPQSX(p,buf,n) {\
    _StructSwapEx(PropertyQueryStructExStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PropertyQueryStructEx));\
}
# define PrepOutgoingXPQS(p,buf,n) {\
    _StructSwapEx(XPropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(XPropertyQueryStruct));\
}
# define PrepOutgoingPRPQS(p,buf,n) {\
    _StructSwapEx(PrpQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,LTOR,n,sizeof(PrpQueryStruct));\
}
# define PrepOutgoingCONQS(c) {\
    _LSWAP(c.contract.EqmSizeIn);\
    _LSWAP(c.contract.EqmSizeOut);\
    _LSWAP(c.pollrate);\
    _LSWAP(c.nconsumers);\
  }
# define PrepIncomingPQS(p,buf,n) {\
    _StructSwapEx(PropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PropertyQueryStruct));\
  }
# define PrepIncomingPQSX(p,buf,n) {\
    _StructSwapEx(PropertyQueryStructExStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PropertyQueryStructEx));\
  }
# define PrepIncomingXPQS(p,buf,n) {\
    _StructSwapEx(XPropertyQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(XPropertyQueryStruct));\
  }
# define PrepIncomingPRPQS(p,buf,n) {\
    _StructSwapEx(PrpQueryStructStruct,(unsigned char *)buf,(unsigned char *)p,RTOL,n,sizeof(PrpQueryStruct));\
  }
# define PrepIncomingHRS(rb,buf) {\
    _StructSwapEx(HRSStructStruct,(unsigned char *)buf,(unsigned char *)rb,RTOL,1,sizeof(HRS));\
  }
# define AssignClnStructDataPtr(dst,src,dir) {\
    if (dir == RTOL) {\
      StructSwapEx(din->dTag,(BYTE *)src,(BYTE *)s->sdata,RTOL,din->dArrayLength,s->size);\
      dst = s->sdata;\
    }\
  }
# define cpyClnStructData(dst,src,siz) {\
    if (din->dFormat == CF_STRUCT && (s=getStructFromTag(din->dTag)) != NULL)\
      StructSwapEx(din->dTag,(BYTE *)dst,(BYTE *)src,LTOR,din->dArrayLength,s->size);\
    else memcpy(conTbl[slot]->dataIn,din->data.vptr,siz);\
  }
# define AssignOutgoingHeader(h,buf) memset(h,0,sizeof(HeaderStruct))
# define AssignOutgoingBrdcstData(h,buf) memset(h,0,sizeof(GlobalDataHdr))
# define PrepOutgoingHeader(h,dat) {\
    _StructSwap(HeaderStructStruct,(unsigned char *)dat,(unsigned char *)h,LTOR);\
  }
# define PrepOutgoingGCastData(h,dat) {\
    _StructSwap(GlobalDataHdrStruct,(unsigned char *)dat,(unsigned char *)h,LTOR);\
  }
# define PrepIncomingGCastData(h,dat) {\
    _StructSwap(GlobalDataHdrStruct,(unsigned char *)dat,(unsigned char *)h,RTOL);\
  }
# define PrepIncomingData(dst,src,dsiz,rsiz,fmt) {\
    if (fmt == CF_STRUCT && c != NULL && (s=getStructFromTag(c->strTagOut))) {\
      StructSwapEx(c->strTagOut,(BYTE *)src,(BYTE *)dst,RTOL,rsiz/s->rawlen,s->size);\
    } else {\
      memcpy(dst,src,MIN(dsiz,rsiz));\
      SwapBytes((char *)dst,rsiz/getFormatSize(fmt),fmt,RTOL);\
    }\
    if ((dsiz) > (rsiz)) memset(&((char *)(dst))[rsiz],0,dsiz-rsiz);\
    if (c != NULL && c->field[0] != 0) mapBitfieldToTarget(c->context,c->expName,c->strTagOut,c->field,fmt,(BYTE *)dst);\
  }
# define PrepOutgoingData(d,s,siz,fmt) {\
    memcpy(d,s,siz);\
    SwapBytes((char *)d,siz/getFormatSize(fmt),fmt,LTOR);\
  }

#endif  /* INTEL (data prepartion MACROS) */

/* client-side macros : */
#define IncrementConnections(n) (n)++

#define CheckSubscriptions(e) \
   while (prepSubRequest(clnWorkArea,e) >= 0);

#define WaitForSyncCount(i) clientCycle()

#if !defined(DOS)
# define initClnWorkArea(i) \
  ( (clnWorkArea=(BYTE *)SystemCalloc(1,clnWorkAreaSize=i)) == NULL ?\
     out_of_client_memory : 0 )
#endif
#ifdef VXWORKS              /* VXWORKS */
  void tineClientCycle(void);
# define vxwStartClientCycle(c) \
   (taskSpawn(c,100,0,10000,tineClientCycle,0,0,0,0,0,0,0,0,0,0) == ERROR ?\
    not_running : 0)
#endif

#define CompIncomingData(out,cmp,sz,f,t) (\
     cmp == NULL ? FALSE :\
     t   != NULL ? !_datcmp(out,cmp,sz,f,t) :\
     !memcmp(out,cmp,sz)\
  )

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* PROLOG_H_ */
