/** \file srvcore.h
 *
 * server-side (and client-side) transport definitions and prototypes.
 *
 */
#ifndef SRVCORE_H_
#define SRVCORE_H_

#include "project.def"
#include "prolog.h"
#include "tinetype.h"
#include "errors.h"
#include "almlib.h"
#include "threader.h"

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifndef MULTITHREADED
#define ReleaseSystemMutex(m) (m = 0)
#define WaitForSystemMutex(m,to) ((m++) > 0 ? -1 : 0)
extern int hSystemClientMutex;
extern int hSystemServerMutex;
extern int hSystemKernelMutex;
extern int hLinkTblMutex;
extern int hLinkQueueMutex;
#else
void ReleaseSyncLink(int flag);
#endif
int WaitForSyncLink(int to);
int InitSystemMutexSet(void);

#define DEFAULT_TRANSPORT_PROTOCOL_LEVEL 6

/**
 * \internal
 * \brief Default value for MaxNumConnections.
 */
#ifndef MAX_ENTRIES
# define MAX_ENTRIES 1000 /* connection table entries  */
#endif

/**
 * \internal
 * \brief Communication protocol macro definitions ...
 */
#define IPX             0x01
#define SPX             0x02
#define TCP             0x04
#define UDP             0x08
#define PIPE            0x10
#define MMF             0x20

#define getProtocolAsString(i) (\
  (i) == IPX ? "IPX" :\
  (i) == SPX ? "SPX" :\
  (i) == TCP ? "TCP" :\
  (i) == UDP ? "UDP" :\
  (i) == PIPE ? "LCL" :\
  (i) == MMF ? "LCL" :\
  "---"\
)

#define CLN_REQ_ 0x01
#define SRV_RSP_ 0x02

/**
 * \internal
 * \brief Tine Engine operation mode
 *
 * Single process OSes (DOS, NIOS) use POLLING, where in some sense, TINE becomes the OS.  
 * Multi-process OSes (UNIX, WINDOWS, VxWorks, VMS, etc.) use SELECTING, where the
 * TINE engines only wakes up if 'select()' fires (incoming data) or if the 
 * SystemPolling rate so dictates.
 */
#define POLLING         0x01
#define SELECTING       0x02

/**
 * \internal
 * \brief The default value for MaxPollingRate.
 *
 * Client contracts asking to poll properties at an interval greater than MaxPollingRate will
 * be automatically converted to MaxPollingRate.
 * This in effect sets the upper limit for the timeout parameter for select() when SELECTING
 * mode is in effect.  Note that 'Rate' is actually a misnomer as the value of this parameter
 * actually gives an interval and not a rate.
 *
 */
#ifndef DEFAULT_POLLRATE
#  ifndef SYSPOLL
#    ifdef MULTITHREADED
#      define DEFAULT_POLLRATE 20
#    else
#      define DEFAULT_POLLRATE  1000
#    endif
#  else
#    define DEFAULT_POLLRATE SYSPOLL
#  endif
#endif

/**
 * \internal
 * \brief The default value for MinPollingRate.
 *
 * Client contracts asking to poll properties at an interval smaller than MinPollingRate will
 * be automatically converted to MinPollingRate.  Thus a renegade client cannot force a server
 * to poll a property at smaller intervals.
 *
 */
#ifndef MIN_POLLRATE
#  ifndef SYSPOLL
#    define MIN_POLLRATE 20
#  else
#    define MIN_POLLRATE SYSPOLL
#  endif
#endif

/**
 * \internal
 * \brief The default timeout used when ExecLink() is called.
 *
 * ExecLink() calls ExecLinkEx() with DEFAULT_CLIENT_POLLRATE as the timeout parameter.
 *
 */
#define DEFAULT_CLIENT_POLLRATE 1000

/**
 * \internal
 * \brief The default value for MaxSystemTransportSize
 *
 * Buffer space is reserved at initialization according to this value (64 KBytes).  
 * Servers which need to support larger data transfers should set this value
 * accordingly.
 *
 * \todo Make this gracefully dynamic if possible. (Watch out for VxWorks, NIOS).
 *
 */
#define MAX_TRANSPORTSIZE 0xffff

 /**
 * \internal
 * \brief The default value of gBurstLimit.
 *
 * This corresponds to the number of data packets sent without a break.
 * If UDP is in use, there is no flow control, so data transfers are made assuming that
 * all clients can handle the network speed of the server.  
 * This may or may not be true. The default value is a generally reasonable choice, but
 * Video servers sending large blocks of data will probably want to set gBurstLimit to a 
 * high value on initialization.
 *
 */
#ifndef BURST_LIMIT
# define BURST_LIMIT 20
#endif
extern int gBurstLimit;

 /**
 * \internal
 * \brief The default value of gCycleDelay.
 *
 * This corresponds to the number of milliseonds to delay following a burst transmission.
 * If UDP is in use, there is no flow control, so data transfers are made assuming that
 * all clients can handle the network speed of the server.  
 * This may or may not be true. The default value is a generally reasonable choice, but
 * Video servers sending large blocks of data will probably want to set gCycleDelay to a 
 * smaller value (or 0) on initialization.
 *
 */
#ifndef CYCLE_DELAY
# define CYCLE_DELAY   20
#endif
extern int gCycleDelay;

/**
 * \internal
 * \brief The default value of gSystemTick.
 *
 * Used internally within the TINE engine scheduler.
 *
 */
#ifndef SYSTEM_TICK
# define SYSTEM_TICK 10
#endif
extern UINT32 gSystemTick;

/**
 * \internal
 * \brief The default value of gSynchronizeContracts.
 *
 * Determines whether the TINE engine scheduler should attempt to
 * force contracts with polling intervals of (modulo) 1000 msec
 * should be called consecutively within the same cycle.  This
 * greatly improves the 'packing' of contracts being returned to
 * a given client.
 *
 */
#ifndef CONTRACT_SYNCHRONIZE
# define CONTRACT_SYNCHRONIZE TRUE
#endif
extern UINT32 gSynchronizeContracts;

/**
 * \internal
 * \brief Used as a flag in lieu of the startup time field in a contract.
 *
 * A server servicing a NETWORK subscription will set the startup time 
 * field in the contract header to BCAST_ID (for either a multicast or
 * broadcast request) as a special signature.
 *
 */
#define BCAST_ID 255

/**
 * \internal
 * \brief Special scheduling signals for the TINE scheduler.
 *
 */
#define PRP_SCHEDULE_SIGNAL 0
#define PRP_RETRY_SIGNAL 1
#define PRP_CANCEL_SIGNAL 2

/**
 * \internal
 * \brief Special reserved system names ...
 *
 */
#define SRVEQM_NAME "_SRV__"
#define SRVFEC_NAME "NETWORK"
#define SRVEXP_NAME "NETWORK"
#define STKEQM_NAME "_STK__"
/**
 * \internal
 * \brief Determine whether this request is coming from a NETWORK request
 * for services (e.g. ENS, CAS, etc.).
 *
 */
#define isServiceRequest(pctr) (strcmp((pctr)->EqmName,SRVEQM_NAME) == 0 ? TRUE : FALSE)

/**
 * \internal
 * \brief The default value for gRespondToServiceRequests.
 *
 * Determine whether this server can respond to Service requests.
 * Typically only the ENS will have this features turned on.  (Or possibly the CAS) 
 *
 */
#ifndef RESPOND_TO_SERVICE_REQUESTS
# define RESPOND_TO_SERVICE_REQUESTS FALSE
#endif
extern TINE_EXPORT int gRespondToServiceRequests;

#define REGISTERED_ENS  0x01
#define REGISTERED_GENS 0x02
#define REGISTERED_ALL  0x03

typedef struct EQPTask
{
  void (*tskfcn)(void); /* background task function */
  int weight;           /* background task weight factor */
  struct EQPTask *next;
} EQPTASK;
extern EQPTASK *EQPTaskList;

#define EQPBKG void (*)(void)
#define EQPINI void (*)(void)
#define EQPEXI void (*)(void)

#define ALLOCSIZE(b) ((b) + (8 - (b) % 8))

/**
 * \internal
 * \brief The top level header which defines the incoming data packet.
 * 
 * \todo split TotalSize away from the header.  This will involve a good look
 * at the routines in swaplib.c.  The advantage of doing this is primarily to 
 * deal with a HeaderStruct with fields on a 4-byte boundary.  The first two
 * bytes of what has come in doesn't need to be a part of this header. BUT,
 * splitting this off from the header has low priority.
 *
 */
typedef struct       /* struct is passed over the net */
{

  /** The total size in bytes (as a consistency check,
   * must match the bytes read) of everthing in this packet.
   */
  UINT16 TotalSize;

  /** The Client name sending the contract. */
  char UserName[USERNAME_SIZE]; 

  /** The contract header protocol level (levels 1 and 2
   * are no longer supported; 4 is probably no longer in use; 3 is
   * used by globals contracts; 5 is currently modern).
   */
  short Protocol;
  
   /**  A vestige field which nominally should give the number
    * of independent contract headers to follow. (Not used in protocol level 5).
    */
  short EqmSize;

} HeaderStruct;
#define HEADERSTRUCT_SIZE  22
#define SUB_PROT_LOCATION 18

/**
 * \internal
 * \brief Linked List structure used by clients to manage redirected calls.
 */
typedef struct RedirectedLinkStruct
{
  char srcContext[CONTEXT_NAME_SIZE];
  char srcServerName[EXPORT_NAME_SIZE];
  char srcProperty[PROPERTY_NAME_SIZE];
  char srcDeviceName[DEVICE_NAME_SIZE];
  char dstContext[CONTEXT_NAME_SIZE];
  char dstServerName[EXPORT_NAME_SIZE];
  char dstDeviceName[DEVICE_NAME_SIZE];
  char dstProperty[PROPERTY_NAME_SIZE];
  struct RedirectedLinkStruct *prv;
  struct RedirectedLinkStruct *nxt;
} REDIRECTED_LINK;
extern REDIRECTED_LINK *RedirectedLinkList;

/**
 * \internal
 * \brief Linked List structure used by connection table if multiple links to same contract
 */
typedef struct DependentLinkStruct
{
  int linkId;
  struct DependentLinkStruct *prv;
  struct DependentLinkStruct *nxt;
} DEPENDENT_LINK;

typedef struct
{
  float absolute;
  float percent;
} DataToleranceFactor;

/**
 * \internal
 * \brief Linked List structure used by clients to manage grouped calls.
 */
typedef struct grpStruct
{
  short numPending;     /* num of group events pending */
  short numInGrp;       /* number of elements in group */
  size_t GrpID;         /* group identification */
  short  GrpCc;         /* group completion */
  struct grpStruct *next;
} GroupTableEntry;

#ifdef DOS
# define WILDCARD_LINKID(i) ((i) & 0xff)
# define WILDCARD_LISTID(i) ((i) >> 8)
# define MAKE_WILDCARD_CALLBACKID(wcid,lstid) ((wcid) + ((lstid)<<8))
#else
# define WILDCARD_LINKID(i) ((i) & 0xffff)
# define WILDCARD_LISTID(i) ((i) >> 16)
# define MAKE_WILDCARD_CALLBACKID(wcid,lstid) ((wcid) + ((lstid)<<16))
#endif

/**
 * \internal
 * \brief Table entry structure used by clients to manage wildcard calls
 */
typedef struct wildCardTableStruct
{
  int linkId;  /* xref into connTableEntry */
  int callbackId; /* user's callback id */
  void (*callback)(int,int); /* user's callback */
  int length;  /* length of device list */
  int format;  /* original caller format */
  int numPending; /* number waiting to notify */
  int numActive;  /* number of active links in this set */
  NAME16 *list;/* list of device names */
  BYTE *data;  /* either floats or longs */
  int *status; /* list of individual comp codes */
  int *listId; /* list of individual link ids */
} wildCardTableEntry;

extern wildCardTableEntry **wcTable;
extern int nWildCardTableEntries;

#define UNASSIGNED_CALLBACKID 0xeeeeeeee
#define DEFAULT_RECV_QUEUE_DEPTH 10
typedef struct ConnTableQueueStruct
{
  BYTE *dataOut;
  int dataSize;
  int dataStatus;
  int statusCode;
  UINT32 timeStamp;
  int timeStampMSEC;
  int usrStamp;
  int sysStamp;
  int xfrReason;
  struct ConnTableQueueStruct *next;
} ConnTableQueue;
/**
 * \internal
 * \brief Connection table structure used by clients.
 */
typedef struct          /* local */
{              
  char expName[EXPORT_NAME_SIZE]; /* eqm server export name */
  char prpName[PROPERTY_NAME_SIZE]; /* eqm property */
  char devName[DEVICE_NAME_SIZE]; /* eqm device name */
  char context[CONTEXT_NAME_SIZE]; /* eqm context */
  char *extStringspace; /* eqm extended string space */
  short extStringsize;
  short access;         /* data access */
  short formatIn;       /* eqm data format down */
  short formatOut;      /* eqm data format up */
  UINT32 sizeIn;        /* data array size down */
  UINT32 sizeOut;       /* data array size up */
  char strTagIn[TAG_NAME_SIZE]; /* eqm structure tag down */
  char strTagOut[TAG_NAME_SIZE]; /* eqm structure tag up */
  char field[16];       /* target bitfield field or structure field */
  short dataStatus;     /* signal new arrival (negative completion code) */
  short tocounter;      /* link timeout counter */
  int  srvIdx;          /* index into Srv Table */
  int  fecIdx;          /* index into FEC database */
  UINT32 pollRate;      /* polling rate for connectioned clients */
  UINT32 heartbeat;     /* used by subscriptions */
  UINT16 mode;          /* used by subscriptions */
  UINT16 xferReason;    /* transfer flag applicable to last data transfer */
  int tineProtocol;     /* tine protocol level */
  int inetProtocol;     /* transport protocol (IPX,SPX,UDP,TCP) */
  UINT32 time;          /* last polling request */
  UINT32 starttime;     /* entry into table (used by subscriptions) */
  UINT32 dtimestamp;    /* arriving TINE timestamp */
  int dtimestampMSEC;   /* arriving TINE timestamp (msec fraction) */
  int datastamp;        /* call applied data stamp */
  int systemstamp;      /* system applied data stamp */
  UINT32 lastDataSize;  /* arriving data size */
  int  TCPsocket;
  UINT32 mcastGrp;      /* the multicast group ip addr (if attached to a group) */
  int  counter;         /* current subscription counter */
  char *dataIn;         /* pointer to associated data down */
  char *dataOut;        /* pointer to associated data up */
  char *bufOut;         /* pointer to raw data up buffer if necessary */
  BYTE *bufOutBlks;     /* pointer to data-up buffer block progress if necessary */
  char *bufCmp;         /* pointer to comparison buffer if supressHearbeatNotification == TRUE */
  DataToleranceFactor *dtf; /* pointer to comparison tolerance if applied */
  short allocated;      /* is dataOut allocated memory ? */
  short useErrValue;    /* flag  */
  UINT16 numblks;       /* progress counters for incoming data */
  UINT16 blksin;        /* progress counters for incoming data */
  UINT32 bytesin;       /* progress counters for incoming data  */
  UINT32 nextDataSize;  /* next outgoing data size in bytes */
  UINT32 nextDataPtr;   /* next outgoing data array pointer */
  UINT32 bytesRemaining;/* number of bytes data remaining */
  UINT32 bytesTotal;    /* total number of bytes data */
  UINT16 datamtu;       /* outgoing data minimum transport unit */
  UINT16 blkidIn;       /* block Id of data to send */
  UINT16 numblksIn;     /* num blks input data to send */
  UINT16 blknumIn;      /* this blk num of input data */
  UINT32 callbackID;    /* pivate callback ID */
  void (*fcn)(int id,int cc); /* call back function */
  UINT8 grouped;        /* flag to indicate if element is grouped */
  UINT8 notifyPending;  /* flag to indicate if notification is pending */
  UINT8 supressCallback;/* flag to withhold callback notification */
  UINT8 reserved; 
  UINT16 srvID;         /* server subscription ID (applicable in MCAST) */
  UINT32 rdrID;         /* if non-null gives child connection ID for redirected calls */
  wildCardTableEntry *wcLink; /* wildcard entry if relevant */
  short linkStatus;     /* current link status */
  short needsNotification; /* flag to call NotifyClient() or not */
  short hasNotifiedOnce; /* necessary for addressing nofication suppression issues */
  short hasUserCallbackId; /* flag : TRUE if set by caller, otherwise false */
  short lastLnkErrSrc; /* LCL or RMT */
  char lastStatusString[STATUS_SIZE]; /* temporary storage for return status */
  BYTE  errValue[64];   /* buffer for assigned error value */
  REDIRECTED_LINK *rdrLink; /* pointer into the redirection table */
  int cancelPending;    /* delay CANCEL because of dependencies */
  int boundToId;        /* id of parent link if this one is bound (>0) */
  int mcaIndex;
  DEPENDENT_LINK *dpdLink; /* pointer to dependency list (if this link is a parent) */
  int queueLen;
  int queueDepth;
  ConnTableQueue *queue;
} ConnTableEntry;

typedef struct          /* local */
{              
  char expName[EXPORT_NAME_SIZE]; /* eqm function name */
  char prpName[PROPERTY_NAME_SIZE]; /* eqm property */
  char devName[DEVICE_NAME_SIZE]; /* eqm device name */
  char context[CONTEXT_NAME_SIZE]; /* eqm context */
  char eqm[EQM_NAME_SIZE];
  UINT16 mode;          /* used by subscriptions */
  short access;         /* data access */
  short formatIn;       /* eqp data format down */
  short formatOut;      /* eqp data format up */
  UINT32 sizeIn;        /* data array size down */
  UINT32 sizeOut;       /* data array size up */
  char strTagIn[TAG_NAME_SIZE]; /* eqm structure tag down */
  char strTagOut[TAG_NAME_SIZE]; /* eqm structure tag up */
  UINT32 pollRate;      /* polling rate for connectioned clients */
  UINT32 heartbeat;     /* used by subscriptions */
  UINT32 starttime;     /* entry into table (used by subscriptions) */
  UINT32 dtimestamp;    /* arriving RPC timestamp */
  int dtimestampMSEC;   /* arriving RPC timestamp (msec fraction) */
  int  counter;         /* current subscription counter */
  short linkStatus;     /* current link status */
} ConTblInfo;
TINE_EXPORT int GetConnectionTable(ConTblInfo *tbl,int *tblSize);

#define EMPTY_PROP    0
#define TARGET_PROP   1
#define WILDCARD_PROP 2
#define STOCK_PROP    3

#ifndef PROPERTY_HASH_SIZE
# define PROPERTY_HASH_SIZE 211
#endif
#define PROPERTY_HASH_INDEX(i) ((i) & 0xff)
#define PROPERTY_NUM_INDEX(i) (((i) >> 8) & 0xffff)
#define MAKE_PROPERTY_ID(i,indx) (((i) << 8) | (indx))

#define PRP_REDIR_CTX_OFFSET 0
#define PRP_REDIR_SRV_OFFSET 32
#define PRP_REDIR_DEV_OFFSET 64
#define PRP_REDIR_PRP_OFFSET 128

#define PRP_REDIR_SRV_OFFSET_P5 0
#define PRP_REDIR_PRP_OFFSET_P5 16

#ifndef DEVICES_HASH_SIZE
# define DEVICES_HASH_SIZE 211
#endif
#define DEVICES_HASH_PAYOFF (DEVICES_HASH_SIZE/10) /* don't hash if fewer modules than this */

typedef struct HashXRef
{
  int index;
  struct HashXRef *next;
} HashXRefStruct;

#ifndef STOCKPROP_HASH_SIZE
# define STOCKPROP_HASH_SIZE 103
#endif

extern HashXRefStruct **StockPropXRefTable;

typedef struct 
{
  NAME64 *prpList;
  int prpListSize;
  char filename[64];
} EqmDevicePropertyList;

/**
 * \internal
 * \brief export info struct (to handle file input)
 *
 */
typedef struct ExpInfoStruct
{
  char EqmContext[CONTEXT_NAME_SIZE]; /* exported equipment name context */
  char ExpName[EXPORT_NAME_SIZE]; /* exported equipment module name */
  char SubSystem[SUBSYSTEM_NAME_SIZE]; /* equipment module sub system */
  char EqmName[EQM_NAME_SIZE]; /* local equipment module name */
  char EqmGroup[EXPORT_NAME_SIZE]; /* group name, if applicable */
  char PrpName[PROPERTY_NAME_SIZE];
  char fmtOut[64];
  char fmtIn[64];
  char description[128];
  char redirection[DEVICE_REDIR_SIZE];
  char access[16];
  int  PrpSizeOut;
  int  PrpSizeIn;
  int  PrpId;
  int  EqmGroupIdx;        /* group index if applicable */
  int  DevSpace;/* total amount of space initially allocated */
} XIS;

typedef struct ContractResourceStruct
{
  DTYPE din;
  DTYPE dout;
  char devName[DEVICE_NAME_MAXSIZE];
  char devProp[PROPERTY_NAME_SIZE+4];
  short access;
} CRS;

typedef struct 
{
  char *rdrName;
} EqmDeviceRdrListType;
/**
 * \internal
 * \brief Linked list structure used to hold equipment module information.
 *
 */
typedef struct ExportListTag /* local */
{
  char   EqmContext[CONTEXT_NAME_SIZE]; /* exported equipment name context */
  char   EqmExportName[EXPORT_NAME_SIZE]; /* exported equipment module name */
  char   EqmSubSystem[SUBSYSTEM_NAME_SIZE]; /* equipment module sub system */
  char   EqmName[EQM_NAME_SIZE]; /* local equipment module name */
  char   EqmAlias[EQM_NAME_SIZE]; /* local equipment alias name */
  char   EqmGroup[EXPORT_NAME_SIZE]; /* group name, if applicable */
  int    groupIndex;        /* group index if applicable */
  int    exportId;          /* local ID */
  ExportPropertyListStruct  /* pointer to chained equipment property hash list */
         *EqmProperty[PROPERTY_HASH_SIZE+1]; /* Add room for Wild-card list */
  int    (*EqmPropertyQueryFcn)(char *devName,PrpQueryStruct **prpqs);
  int    (*XEqmPropertyQueryFcn)(char *devName,XPropertyQueryStruct **xpqs);
  short   hasPropertyDeviceNames;
  short   hasRegisteredDevicesNames;
  NAME64 *EqmDeviceList;    /* pointer to device name list */
  EqmDeviceRdrListType *EqmDeviceRdrList; /* pointer to device redirection list (context/server inside a single NAME64) */
  EqmDevicePropertyList *EqmDevicePrpList;/* pointer to array of device-property lists */
  HashXRefStruct **EqmDeviceXRef; /* ModuleName cross-reference */
  short  EqmNumProperties;  /* total number of properties supported */
  short  EqmNumDevices;     /* total number of devices supported */
  short  EqmDeviceSpaceSize;/* total amount of space initially allocated */
  short  registered;        /* registered with name server */
  ADSList *adsTable;        /* alarm definition structure table (linked List) */
  short  adsTableSize;      /* number of alarm definitions */
  short  alarmSystem;       /* alarm system flag */
  ALARM  **almLst;          /* associated alarm list */
  short  nalarms;           /* number of active alarms */
  UINT32 almMRT;            /* alarm Most Recent Timestamp */
  UINT32 almMRCT;           /* alarm Most Recent Complete Timestamp */
  short  almHSV;            /* alarm highest Severity */
  short  nAlmMRT;           /* number of alarms at MRT */
  short  nAlmMRCT;          /* number of alarms at MRCT */
  short  nAlmHSV;           /* number of alarms at HSV */
  short  nclients;          /* number of active clients */
  short  ncontracts;        /* number of active contracts */
  int  (*eqmFcn)(char *,char *,DTYPE *,DTYPE *,short);/* associated equipment function */
  void (*eqmBkg)(void);     /* associated background task */
  int hEqmBkg;              /* handle to background task (in case of threads) */
  SINT32 rateBkg;           /* background polling rate msec */
  short bkgbusy;            /* background busy semaphore */
  short inidone;            /* init task done flag */
  short hstdone;            /* history init done flag */
  short eqmbusy;            /* eqm fcn call is busy */
  struct timeval lastBkg;   /* background task last polled */
  void (*eqmIni)(void);     /* associated initialization  */
  void (*eqmExi)(void);     /* associated cleanup */
  char EqmLastErrorString[STATUS_SIZE];
  int EqmLastErrorCode;
# ifdef MULTITHREADED
  CRS EqmCrs;
# ifdef UNIX
  pthread_mutex_t EqmCallMutex[2];
  pthread_mutex_t DataProtectionMutex;
# endif
  MXHANDLE hEqmCallMutexPair[2];
  MXHANDLE hDataProtectionMutex;
  int EqmCallMutexIndex;
  int hEqmCallThread;
# endif
  void *EqmCls;
  struct ExportListTag *next;  /* Linked list */
} ExportListStruct;

extern int gEqmFcnAccessor;
extern ExportPropertyListStruct *ExportPropertyList;
extern ExportListStruct *ExportList;
extern int gEqmListSize;
extern int gSystemUseDataProtection;
TINE_EXPORT void SetSystemUseDataProtection(int value);

/**
 * \internal
 * \brief A Structure which is used to map a EqmContext/ExportName to a FecName/EqmName.
 * 
 * The client uses EqmContext and Export name to address the FEC. 
 * TINE internally uses the FecName and the Equipment name.
 * 
 * The TINE library uses this structure for internal caches.
 */
typedef struct                      /* struct is passed over the net (old, no swapping) */
{
  char  FecName[FEC_NAME_SIZE];     /**< FEC Name */
  char  SubSystem[SUBSYSTEM_NAME_SHORTSIZE]; /* short sub-system name */
  char  ExportName[EXPORT_NAME_SHORTSIZE];	/**< Export Name */
  char  EqmName[EQM_NAME_SHORTSIZE+1];	  /**< Equipment name */
  char  EqmContext[CONTEXT_NAME_SHORTSIZE];	/**< EqmContext */
  short alarmSystem;
} RPCDataStruct;
#define RPCDATASTRUCT_SIZE (FEC_NAME_SIZE+SUBSYSTEM_NAME_SHORTSIZE+EXPORT_NAME_SHORTSIZE+EQM_NAME_SHORTSIZE+1+CONTEXT_NAME_SHORTSIZE+2)

typedef struct                      /* struct is passed over the net (new) */
{
  char  FecName[FEC_NAME_SIZE];
  char  SubSystem[SUBSYSTEM_NAME_SIZE];
  char  ExportName[EXPORT_NAME_SIZE];
  char  EqmName[EQM_NAME_SIZE];
  char  EqmContext[CONTEXT_NAME_SIZE];
} ExpDataStruct;
#define EXPDATASTRUCT_SIZE (FEC_NAME_SIZE+SUBSYSTEM_NAME_SIZE+EXPORT_NAME_SIZE+EQM_NAME_SIZE+CONTEXT_NAME_SIZE)

extern UINT32 IncomingNetworkProtocol;
typedef struct                      /* struct is passed over the net (old, no swapping) */
{
  char Name[FEC_NAME_SIZE];
  BYTE reserved[3];
  BYTE Port[2];
  BYTE Network[4];
  BYTE Node[6];
  BYTE ImmediateAddress[6];
  char IP[16];
  BYTE PortOffset;
  char IPh_addr[4];
  short  Prot;
} RPCFecStruct;
#define RPCFECSTRUCT_SIZE (FEC_NAME_SIZE+3+2+16+21+2)

typedef struct                      /* struct is passed over the net (new) */
{
  char Name[FEC_NAME_SIZE];
  BYTE IPXNetwork[4];
  BYTE IPXNode[6];
  BYTE IPXImmediateAddress[6];
  char IP[ADDR_SIZE];
  char IPh_addr[4];
  SINT32 PortOffset;
  SINT32 TransportProtocol;
  SINT32 TineProtocol;
} FecDataStruct;
#define FECDATASTRUCT_SIZE (FEC_NAME_SIZE+16+16+4+3*4)

typedef struct
{
  char os[12];
  char desc[64];
  char loc[32];
  char hdw[32];
  char sub[16];
  char resp[32];
} RPCFecExStruct;

typedef struct /* new */
{
  char os[FEC_OS_SIZE];
  char desc[FEC_DESC_SIZE];
  char loc[FEC_LOCATION_SIZE];
  char ver[FEC_VERSION_SIZE];
  char hdw[FEC_HDW_SIZE];
  char resp[FEC_RESP_SIZE];
} FecInfoStruct;

typedef struct /* used for holding data temporarily during initialization */
{
  char fecName[FEC_NAME_SIZE];
  char fecOs[FEC_OS_SIZE];
  char fecContext[CONTEXT_NAME_SIZE];
  char fecSubsystem[SUBSYSTEM_NAME_SIZE];
  char fecDescription[FEC_DESC_SIZE];
  char fecLocation[FEC_LOCATION_SIZE];
  char fecHardware[FEC_HDW_SIZE];
  char responsible[FEC_RESP_SIZE];
  char expNameTarget[EXPORT_NAME_SIZE];
  char eqmNameTarget[EQM_NAME_SIZE];
  char historyHome[HISTORY_HOME_SIZE];
  int  fecPort;
} FIDS;

typedef struct 
{
  char devname[DEVICE_NAME_SIZE];
  int  devnum;
  char devrdr[DEVICE_REDIR_SIZE];
  char prplst[32];
} DEVIS;

typedef struct
{
  char FecName[FEC_NAME_SIZE];
  char reserved[4];
  UINT32 localtime;
  UINT32 starttime;
  UINT32 systemPollingRate;
  short  numBkgTsks;
  short  numTotalContracts;
  short  numTotalClients;
  short  numTargetContracts;
  short  numTargetClients;
  short  numConnections;
  UINT32 numConnectionTimeouts;
  UINT32 numConnectionArrivals;
  UINT32 numUDPpkts;
  UINT32 numTCPpkts;
  UINT32 numIPXpkts;
  UINT32 numSPXpkts;
} ActivityQueryStruct;
#define ACTIVITYQUERYSTRUCT_SIZE (FEC_NAME_SIZE+4+4*3+2*6+4*6)

typedef struct
{
  char desc[64];
  BYTE dataFmt;
  BYTE dataSiz;
  BYTE data[64];
  UINT32 timeEntered;
  UINT32 codeNow;
  UINT32 codePrv;
  UINT32 codeNxt;
} ServerStateStruct;
#define SERVERSTATESTRUCT_SIZE (64+2+64+4*4)

typedef struct
{
  UINT32 AveBusyTime;
  UINT32 CycleCounts;
  UINT32 MaxCycleCounts;
  UINT32 SingleLinkCount;
  UINT32 ClientMisses;
  UINT32 ClientReconnects;
  UINT32 ClientRetries;
  UINT32 ContractMisses;
  UINT32 ContractDelays;
  UINT32 BurstLimitReachedCount;
  UINT32 DataTimeStampOffset;
} ServerStatsStruct;
#define SERVERSTATSSTRUCT_SIZE (11*4)

/**
 * \internal
 * \brief Determines the default value of MaxNumClients
 *
 * The places an upper limit on the number of simultaneous clients a server is willing
 * to manage.  If it is known that a server needs to service more clients than this
 * default, then the value of MaxNumClients should be set accordingly at initialization time.
 */
#ifndef NCONSUMER
# define NCONSUMER     100
#endif

/**
 * \internal
 * \brief Determines the default value of MaxNumContracts
 *
 * The places an upper limit on the number of simultaneous contracts a server is willing
 * to manage.  If it is known that a server needs to service more contracts than this
 * default, then the value of MaxNumContracts should be set accordingly at initialization time.
 */
#ifndef NCONTRACT
# define NCONTRACT     1000
#endif

#define RENEWAL       60
#define REMINDER      5
#define ACK_REQUEST   (2*REMINDER)
#define HEARTBEAT     60
#define REPLY_PENDING 3
#define SUB_PENDING   0x3fff
#define ACK_PENDING   0x7fff

/**
 * \internal
 * \brief The payload size of an IPX Datagram.
 *
 */
#define IPX_DATA_SIZE   512
#define IPX_BUFFER_SIZE 1466

/**
 * \internal
 * \brief The payload size of a UDP Datagram.
 *
 * The value here works almost everywhere except NIOS (where #define UDP_BUFFER_SIZE is
 * set to 1024 in project.def).
 *
 * \note TINE performs its own packet reassembly for incoming datagrams.  For many OSes 
 * this is not necessary, but as packet reassembly cannot be assumed to be present or 
 * might have an upper limit relating to the number of recv buffers, it is a fundemental
 * feature of TINE.  
 */
#ifndef UDP_BUFFER_SIZE
# define UDP_BUFFER_SIZE 1472
#endif
/**
 * \internal
 * \brief The default value for one of srvWorkAreaSize, WorkAreaSize, and/or tmpWorkAreaSize.
 *
 * TINE uses a work area buffer for managing contracts calls.  On some (single-threaded) platforms 
 * all work areas point to the same buffer.
 */
#ifndef WORK_SIZE
# define WORK_SIZE  0x10000 /* 64 KB */
#endif

/**
 * \internal
 * \brief The default value for one of clnWorkAreaSize.
 *
 * TINE uses a work area buffer for managing client calls.  On some (single-threaded) platforms 
 * this points to the same area as the srvWorkAreaSize.
 */
#ifndef MIN_WORK_SIZE
# define MIN_WORK_SIZE 2000
#endif

/**
 * \internal
 * \brief The maximum payload size for incoming network globals.
 *
 */
#define BRDCST_BUFFER_SIZE IPX_BUFFER_SIZE

/**
 * \internal
 * \brief The buffer size for a console command
 *
 */
#ifndef CMDSIZE
# define CMDSIZE 256
#endif

extern char *SubType[];
extern int nconsumers;
extern int ncontracts;
extern int StaleData;
extern int nExpiredContracts;

#define isWildcardContract(s) (strchr(s,'*') != NULL)
int EqmFcnWithWildcardDevice(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access);
int EqmFcnWithWildcardProperty(ExportListStruct *el,char *DevName,char *DevProperty,DTYPE *dout,DTYPE *din,short access);

/**
 * \internal
 * \brief Contract input header for protocol level 5.
 *
 */
typedef struct             /* struct is passed over the net  */
{
  char   EqmProperty[PROPERTY_NAME_SIZE];
  char   EqmDeviceName[DEVICE_NAME_SIZE];
  char   EqmName[EQM_NAME_SIZE];
  UINT32 EqmSizeIn;
  UINT32 EqmSizeOut;
  BYTE   hEqmName;
  BYTE   EqmAccess;
  BYTE   EqmFormatIn;
  BYTE   EqmFormatOut;
  char   strTagIn[TAG_NAME_SIZE];
  char   strTagOut[TAG_NAME_SIZE];
} CONTRACT;
#define CONTRACT_SIZE (PROPERTY_NAME_SIZE+DEVICE_NAME_SIZE+EQM_NAME_SIZE+4*2+4*1+2*TAG_NAME_SIZE)

typedef struct
{
  CONTRACT contract;
  SINT32 pollrate;
  SINT32 nconsumers;
} ContractQueryStruct;
#define CONTRACT_QUERY_SIZE (CONTRACT_SIZE + 4 + 4)

/* Protocol level 5  : TINE 3.0 - 3.31 */
typedef struct             /* struct is passed over the net  */
{
  char   EqmProperty[PROPERTY_NAME_SHORTSIZE];
  char   EqmDeviceName[DEVICE_NAME_SHORTSIZE];
  char   EqmName[EQM_NAME_SIZE];
  UINT32 EqmSizeIn;
  UINT32 EqmSizeOut;
  BYTE   hEqmName;
  BYTE   EqmAccess;
  BYTE   EqmFormatIn;
  BYTE   EqmFormatOut;
  char   strTagIn[TAG_NAME_SHORTSIZE];
  char   strTagOut[TAG_NAME_SHORTSIZE];
} CONTRACT5;
#define CONTRACT5_SIZE (PROPERTY_NAME_SHORTSIZE+DEVICE_NAME_SHORTSIZE+EQM_NAME_SIZE+4*2+4*1+2*TAG_NAME_SHORTSIZE)

typedef struct
{
  CONTRACT5 contract;
  SINT32 pollrate;
  SINT32 nconsumers;
} Contract5QueryStruct;
#define CONTRACT5_QUERY_SIZE (CONTRACT5_SIZE + 4 + 4)

/**
 * \internal
 * \brief Client struct for protocol level 5.
 *
 */
typedef struct                    /* struct is passed over the net (up to the transport field) */
{
  char               userName[USERNAME_SIZE];
  IPXAddress         IPXaddress;  /* if 0 then IP transfer is assumed */
  BYTE               node[6];     /* IPX routing node (Immediate Address) */
  struct sockaddr_in IPaddress;
  short              count;       /* number of registrations */
  short              protocol;    /* target client protocol level */
  short              transport;   /* transport protocol */
  int                sck;         /* communication socket */
} CLN;
#define CLIENT_SIZE (USERNAME_SIZE + 12 + 6 + 16 + 2 + 2 + 2 + sizeof(int))

/**
 * \internal
 * \brief Client query struct for protocol level 5.
 *
 */
typedef struct                    /* struct is passed over the net */
{
  char               userName[USERNAME_SIZE];
  IPXAddress         IPXaddress;  /* if 0 then IP transfer is assumed */
  BYTE               node[6];     /* IPX routing node (Immediate Address) */
  struct sockaddr_in IPaddress;
  short              count;       /* number of registrations */
  short              protocol;    /* target client protocol level */
} CLNQS;

typedef struct
{
  CLNQS cln;
} ClientQueryStruct;
#define CLIENT_QUERY_SIZE (USERNAME_SIZE + 12 + 6 + 16 + 2 + 2)

/**
 * \internal
 * \brief Definitions relevant to an Access Lock
 *
 */
#define LOCK_UNLOCKED 0
#define LOCK_PREEMPTIVE 1
#define LOCK_PERSISTENT 2
#define LOCK_ABORT 3
typedef struct
{
  CLN cln;
  int lockType;
  int lockDuration;
  UINT32 lockToken;
  UINT32 lockStart;
} AccessLockStruct;

/**
 * \internal
 * \brief Client connection status struct used by TINE Engine.
 *
 */
typedef struct                /* local */
{
  short  id;                  /* contract ID as seen by consumer */
  short  mode;                /* SINGLE,REFRESH,POLL,CANCEL */
  SINT32 PollingRate;         /* requested polling rate */
  struct timeval lasttime;    /* time of most recent update */
  UINT32 starttime;           /* start of initial subscription */
  UINT16 counter;             /* counts remaining in subscription(s) */
  UINT16 Stale;               /* > 0 => send new data to consumer */
  UINT16 numblks;             /* total number of message blocks */
  UINT16 blknum;              /* block number of this message */
  short  statusCode;          /* interim completion */
  UINT16 mtu;                 /* maximal transport unit */
  UINT32 nextDataSize;        /* next outgoing data size in bytes */
  UINT32 nextDataPtr;         /* next outgoing data array pointer */
  UINT32 bytesRemaining;      /* number of bytes data remaining */
  UINT32 bytesTotal;          /* total number of bytes data */
  UINT32 misses;              /* data return after expiration */
  UINT32 reconnects;          /* client unexpectedly re-establishes link */
  int    isRetry;             /* the contract is being resubmitted by the client */
} ClientSTATUS;

#define MAXLONGGRAMS 5
typedef struct
{
  UINT32  haddr;               /* identify client by IP address */
  UINT16 SubscriptionID;      /* supplied by the consumer upon registration */
  UINT16 blksin;              /* # blocks arrived from long datagram */
  UINT16 blkid;               /* block id of long datagram */
  UINT8  busy,
	 reserved;
  UINT8  *data;
  UINT32 blktime;
  UINT32 SubsStarttime;
} ContractProgress;

typedef struct ConStruct
{
  CLN              *client;
  ClientSTATUS     status;
  struct ConStruct *next;
} CONSUMER;

#define CONTRACT_ACTIVE    0x00
#define CONTRACT_EXPIRED   0x01
#define CONTRACT_AVAILABLE 0x02

/**
 * \internal
 * \brief Contract List Table.
 *
 */
typedef struct                  /* local */
{
  CONSUMER *consumer;           /* assciated consumer list */
  short    nconsumers;          /* number of consumers for this contract */
  short    contractID;          /* local contract ID */
  CONTRACT contract;            /* see above */
  char     *extStringSpace;     /* extended string buffer */
  UINT32   extStringSize;       /* extended string size */
  char     *lngDeviceName;      /* long device name buffer */
  char     *lngPropertyName;    /* long property name buffer */
  SINT32   PollingRate;         /* in mseconds, smallest number wins ... */
  struct   timeval lasttime;    /* most recent time the data polled */
  struct   timeval lastmarked;  /* last time the data were marked STALE */
  struct   timeval starttime;   /* start time of the current call */
  UINT32   dtimestamp;          /* data timestamp */
  int      dtimestampMSEC;      /* data timestamp (msec fraction) */
  int      datastamp;           /* user data stamp */
  BYTE     *DataOut;            /* data to send  */
  BYTE     *DataIn;             /* data received */
  BYTE     *DataBuf;            /* data working area */
  DTYPE    *dout;               /* prepared output DTYPE */
  DTYPE    *din;                /* prepared input DTYPE */
  UINT32   dataSizeOut;         /* output data size < contract.dataSize */
  int      isStkCal;            /* is a stock-property call ? */
  int      callPending;         /* eqm or stk call is pending (multi-threaded?) */
  int      deliveryPending;     /* pending flag for doDelivery() vs doScheduler() */
  int      expired;             /* expired flag for contract removal */
  int      compStatus;          /* completion status */
  int      returnCode;          /* contract completion return code */
  UINT32   delays;              /* delayed scheduling due to deliveryPending */
  UINT32   misses;              /* missed scheduling (i.e. > 2x polling rate) */
  char     compString[STATUS_SIZE];/* completion string (redirection ?) */
  ExportListStruct *exp;        /* associated export element */
  int      listId;              /* cross-reference to this entry */
} ContractListStruct;

/**
 * \internal
 * \brief Contract returned header struct
 *
 */
typedef struct                  /* struct is passed over the net : Protocol 5 (out) */
{
  UINT16 msgsize;               /* message size in bytes */
  UINT16 SubscriptionID;        /* supplied by the consumer upon registration */
  UINT16 CompletionCode;        /* completion or interim status */
  UINT16 numblks;               /* total number of message blocks */
  UINT16 blknum;                /* block number of this message */
  UINT16 mtu;                   /* maximum transport unit */
  UINT16 timestampMSEC;         /* msec fraction of data timestamp */
  BYTE   EqmFormat;             /* Eqp Format of returned data */
  BYTE   counter;               /* tell consumer how much is left */
  UINT32 timestamp;             /* data timestamp */
  UINT32 ClnStarttime;          /* supplied by the consumer upon registration (@ byte 20)*/
} ProducerHeaderStruct5;
#define PRODUCER5_SIZE (7*2+2*1+2*4)

#define CX_NULL      0x00
#define CX_RESPONSE  0x01
#define CX_STALE     0x02
#define CX_HEARTBEAT 0x04
#define CX_EVENT     0x08
#define CX_TIMER     0x10
#define CX_MESSAGE   0x20
#define CX_ERROR     0x40
#define CX_ATTACH    0x80
#define CX_RETRY    0x100

typedef struct                  /* struct is passed over the net : Protocol 6 (out) */
{
  UINT16 msgsize;               /* message size in bytes */
  UINT16 SubscriptionID;        /* supplied by the consumer upon registration */
  UINT16 CompletionCode;        /* return completion code from the eqm */
  UINT16 numblks;               /* total number of message blocks */
  UINT16 blknum;                /* block number of this message */
  UINT16 mtu;                   /* maximum transport unit */
  UINT16 EqmFormat;             /* Eqp Format of returned data */
  UINT16 counter;               /* tell consumer how much is left */
  UINT16 SrvProtocol;           /* server protocol */
  UINT16 xferReason;            /* transfer reason flag (one of the CX_ codes above) */
  UINT32 ClnStarttime;          /* supplied by the consumer upon registration (@ byte 20)*/
  UINT16 stssize;               /* size of error/status string space */
  UINT16 stscode;               /* access status code */
  UINT32 timestamp;             /* data timestamp */
  UINT32 timestampUSec;         /* usec fraction of data timestamp */
  UINT32 userstamp;             /* application settable data stamp */
  UINT32 sysstamp;              /* systematic data stamp */
} ProducerHeaderStruct;
#define PRODUCER_SIZE (12*2+5*4)
#define PMTU(np,tp)  (((np) == IPX ? IPX_BUFFER_SIZE : UDP_BUFFER_SIZE) - ((tp) == 5 ? PRODUCER5_SIZE : PRODUCER_SIZE) - sizeof(UINT16))
#define PHDR_STTM_OFFSET 20
#define PHDR_PROT_OFFSET 16
#define PHDR_STSSIZ_OFFSET 24
#define PHDR_SUBID_OFFSET 2
#define PHDR_CC_OFFSET 4
#define PDHR_SYST_OFFSET 40
#define PDHR_URST_OFFSET 36

/**
 * \internal
 * \brief Contract header struct
 *
 */
typedef struct subscription  /* struct is passed over the net : Protocol 6 (in) */
{
  UINT16    msgsize;
  UINT16    extsize;        /* extended string space */
  UINT16    mtu;            /* maximum transport unit for incoming data */
  UINT16    numblks;        /* total number of message blocks           */
  UINT16    blknum;         /* block number of this message             */
  UINT16    id;
  short     mode;
  short     blkid;          /* (incremented) block id */
  UINT32    PollingRate;
  UINT32    starttime;
  CONTRACT contract;
} SUBSCRIPTION;

#define SUBSCRIPTION_SIZE (8*2+2*4+CONTRACT_SIZE)
#define SUB_MODE_LOCATION (HEADERSTRUCT_SIZE + 8)

typedef struct subscription5         /* struct is passed over the net : Protocol 5 (in) */
{
  UINT16    msgsize;
  UINT16    mtu;            /* maximum transport unit for incoming data */
  UINT8     numblks;        /* total number of message blocks           */
  UINT8     blknum;         /* block number of this message             */
  UINT16    id;
  short     mode;
  short     blkid;          /* (incremented) block id */
  UINT32    PollingRate;
  UINT32    starttime;
  CONTRACT5 contract;
} SUBSCRIPTION5;

#define SUBSCRIPTION5_SIZE (6*2+2*4+CONTRACT5_SIZE)
#define SUB5_MODE_LOCATION (HEADERSTRUCT_SIZE + 8)
#define SUBMTU(np,tp) (((np) == IPX ? IPX_BUFFER_SIZE : UDP_BUFFER_SIZE) - \
                       ((tp) == 5 ? SUBSCRIPTION5_SIZE : SUBSCRIPTION_SIZE) - \
                       HEADERSTRUCT_SIZE)

extern ContractListStruct **ContractList;
extern CLN **Consumers;         /* registered consumer list */

extern int clnSocketNumber;     /* SPX Client Socket                 */
extern int ProducerSocket;      /* UDP Producer Socket               */
extern int ConsumerSocket;      /* UDP Consumer Socket               */
extern int MCastConsumerSocket; /* UDP MCast/BCast Consumer Socket   */
extern int NetServiceSocket;    /* UDP MCast Net Services Socket     */
extern int GlobalSocket;        /* UDP Global listening Socket       */
extern int SyncSocket;          /* UDP trivial link Socket           */
extern struct sockaddr_in prod_name;
extern AccessLockStruct currentAccessLock;

void clientCycle(void);

typedef struct
{
  char prpName[PROPERTY_NAME_SIZE];
  UINT32 prpSize;          /* maximum supported array size (output) */
  UINT32 prpSizeIn;        /* maximum supported array size (input)  */
  short prpFormat;         /* format (output) */
  short prpFormatIn;       /* format (input) */
  short prpAccessMode;     /* bit 0: read access, bit 1: write access */
  short prpArrayType;      /* output array type */
  char  prpStructTag[TAG_NAME_SIZE]; /* output data structure tag */
  char  prpDescription[64];/* short description of property */
  int  (*fcn)(CONTRACT *,char *,char *);
  int   runInSeparateThread; /* flag for potential slow pokes ... */
} StockPropertyStruct;

extern int NrStockProperties;
extern StockPropertyStruct StockProperty[];
#define GetStockPropertyId(str) \
  getTableIndex(str,StockPropXRefTable,STOCKPROP_HASH_SIZE,(char *)StockProperty,NrStockProperties,OFFSETIN(StockPropertyStruct,prpName),sizeof(StockPropertyStruct))

typedef struct
{
  char modName[16];
  int  (*fcn)(CONTRACT *,char *,char *);
  int  runInSeparateThread; /* flag for potential slow pokes ... */
} StockModifierStruct;
extern int NrStockModifiers;
extern StockModifierStruct StockModifier[];

typedef struct 
{
  CRS StkCrs;
  ContractListStruct *StkCls;
  int isInfoLocked;
  int isThreadBusy;
# ifdef MULTITHREADED
  int  (*fcn)(CONTRACT *,char *,char *);
# ifdef UNIX
  pthread_mutex_t StkCallMutex[2];
# endif
  MXHANDLE hStkCallMutexPair[2];
  int StkCallMutexIndex;
  int hStkCallThread;
# endif
} StockRecourceSet;
extern StockRecourceSet gStockRecourceSet;

short IsMetaProperty(char *Property,char **baseProperty,char **metaProperty);

#ifndef MESSAGE_LOG_DEPTH
#define MESSAGE_LOG_DEPTH 500
#endif
typedef struct 
{
  time_t ts;
  int status;
  char msg[128];
} MsgType;
extern MsgType MsgTbl[MESSAGE_LOG_DEPTH];
extern int gMsgTblPtr;
int msglog(int status,char *text,...);
TINE_EXPORT int GetMsgTable(MsgType *msgtbl,int *nummsg);
TINE_EXPORT void ClearMsgTable(void);

#ifndef ALIAS_HASH_SIZE
# define ALIAS_HASH_SIZE 103
#endif
#define ALIAS_HASH_PAYOFF (ALIAS_HASH_SIZE/10)
typedef struct
{
  char name[DEVICE_NAME_SIZE];
  char alias[DEVICE_NAME_SIZE];
} ALIAS_TABLE;
extern ALIAS_TABLE *gAliasTable;
extern int nAliasTableEntries;
extern HashXRefStruct **AliasXRefTable;

int makeAliasXRefTable(void);
int GetAliasTable(void);
int getTableIndex(char *str,HashXRefStruct **xrf,int hshsiz,char *tbl,int tblsiz,int off,int reclen);
#define getAliasTableIndex(str) \
   getTableIndex(str,AliasXRefTable,ALIAS_HASH_SIZE,(char *)gAliasTable,nAliasTableEntries,OFFSETIN(ALIAS_TABLE,alias),sizeof(ALIAS_TABLE))
#define getAliasTableIndexFromName(str) \
   getTableIndex(str,NULL,0,(char *)gAliasTable,nAliasTableEntries,OFFSETIN(ALIAS_TABLE,name),sizeof(ALIAS_TABLE))

extern TINE_EXPORT void (*cdiSystemInitHook)(void);
extern TINE_EXPORT int (*cdiCloseLinkHook)(int);
extern TINE_EXPORT char *(*getCdiLinkNameHook)(char *,char *);
extern TINE_EXPORT int (*cdiExecLinkHook)(char *,char *,DTYPE *,DTYPE *,short,UINT16);
extern TINE_EXPORT int (*cdiAttachLinkHook)(char *,char *,DTYPE *,DTYPE *,short,int,void (*)(int,int),int,UINT32);
extern TINE_EXPORT void (*cdiSystemFreeHook)(void);

#ifdef SERVER_POLLING_MODE
#define ServerCycleMode    POLLING
#else
#define ServerCycleMode    SELECTING
#endif
extern int RequireAcknowledgments;
extern int AveBusyTime;
extern struct timeval thisCycleTime;
extern BYTE *BrdcstBuffer;
extern UINT32 TCPcumcnt,TCPinitcnt;
extern time_t startuptime;
extern int hConn, hControl;
extern WORD ISwap(WORD w);
extern int TCPloaded,IPXloaded,numErr;
extern char gUserName[20];      /* login name of PC user */
extern TINE_EXPORT char gFecName[20];
extern BYTE *clientbuf;
extern BYTE *WorkArea;
extern BYTE *tmpWorkArea;
extern UINT32 WorkAreaSize;
extern UINT32 tmpWorkAreaSize;

extern char gFecAddrBuffer[30],gClientAddrBuffer[30],fecprot[4];
extern char *SRVDataBuffer;    /* global Buffer pointers for data exchange */
extern int ListenForMCasts;
extern int foregroundTTY;
extern UINT32 CMDcount;

extern ContractListStruct *ContractListItem;
extern ConnTableEntry **conTbl;
extern GroupTableEntry *grpTable;
extern int ReturnToClientCycle;
extern int ReturnToServerCycle;
extern int nConnectionTableEntries;
extern int ExecLinkSemaphore;
extern int AttachLinkSemaphore;
extern int recvNetGlobalSemaphore;
extern struct sockaddr_in srv_name,brdcst_name;
extern int TCPParent,TCPProducerSocket,UDPSocket;
extern struct timeval TCPto, fec_to;
extern int NrOfUsers,WRITEaccess;
extern BYTE rpc_socket[2];
extern const BYTE brdcst_socket[2];
extern BYTE reg_socket[2];
extern const BYTE mcast_socket[2];
extern char PATH[];
extern FecDataStruct *FecTbl, *NameServer;
extern ExpDataStruct *SrvTbl;
extern FecInfoStruct gFecInfo;
extern int numNameServers,curNameServer;
extern int numSrvTblEntries,numFecTblEntries,FecTblSize;
extern UINT32 TCPcount,UDPcount,SPXcount,IPXcount,BRDcount;
extern UINT32 nConnectionTimeouts,nArrivals;
extern UINT32 nPartialTransfers;
extern short gNumEqmBkgTasks;
extern time_t *tnull;
extern TINE_EXPORT char gDeviceContext[20];
extern TINE_EXPORT char gDeviceSubSystem[20];
extern UINT32 gContractMisses;
extern UINT32 gContractDelays;
extern UINT32 gClientMisses;
extern UINT32 gClientReconnects;
extern UINT32 gClientRetries;
extern UINT32 gBurstLimitReachedCount;
extern UINT32 gSingleLinkCount;
extern int gCycleCounts, gMaxCycleCounts;
extern char gLastWriteAccessUsr[USERNAME_SIZE+4];
extern char gLastWriteAccessAddr[ADDR_SIZE+4];
extern char gLastWriteAccessPrp[PROPERTY_NAME_SIZE+4];
extern char gLastWriteAccessDev[DEVICE_NAME_SIZE+4];
extern double gLastWriteAccessTime;
extern int gLastMinuteInitializationDone;
extern int max_rcvbuf;
extern int max_sndbuf;
extern TINE_EXPORT int gRetardSingleContractRemoval;
extern TINE_EXPORT int ServerExitCondition;
extern TINE_EXPORT int ServerInitialized;
extern int acquireDeviceNamesFromProperties;
extern int canSendGlobals;
extern char *ltzname[2];
extern char tExecLinkReturnedCompletion[STATUS_SIZE+4];
extern int tExecLinkReturnedDataSize;
extern int gUsingLegacyAddrCalls;
extern int gTineServicesProtocol;

int prepSubHeader(int protocol,SUBSCRIPTION *pSub,BYTE *pSubstream);
int prepOutgoingSubscription(int protocol,ConnTableEntry *c,char *eqm,int datasize,int subid,BYTE *pSubstream);
int prepIncomingSubscription(int protocol,ProducerHeaderStruct *pPHdr,BYTE *pHdrstream);
int prepProducerHeader(int protocol,int xferReason,CONSUMER *cn,CONTRACT *ct,int returnCode,int dstamp,
                       UINT32 tv_sec,UINT32 tv_msec,UINT16 msgsize,BYTE *pPHdrstream);
#define SUB_STEPSIZE(p) ((p) == 5 ? SUBSCRIPTION5_SIZE : SUBSCRIPTION_SIZE)
#define PHDR_STEPSIZE(p) ((p) == 5 ? PRODUCER5_SIZE : PRODUCER_SIZE)

NAME64 *GetEqpDeviceNames(char *eqm,char *prpname,int num,EqmDeviceRdrListType **rdrnames,EqmDevicePropertyList **prplist);
short IsIllegalProperty(char *eqmName,char *prpName,char *dev,CLN *cln,short access,ExportListStruct **listEntry,char *stsString);
int doContract(ContractListStruct *cl);
void doGCast(struct timeval *thisTime);
int GetPrpId(char *eqm,char *prpName);
void devNamesInitialization(void);
void ExitEQPModules(void);
void RegisterEQPtasks(char *name,void (*ini)(void),void (*tsk)(void),int rate,void (*exi)(void));
void checkTasklist(void);
int GetDeviceListFromProperty(char *eqm,char *prp,NAME64 *devlst,UINT32 *devlen);
ExportPropertyListStruct *AssignDeviceListFromProperty(char *eqm,char *prp);
int prepSubRequest(BYTE *data,int isENS);
int sendSubRequest(int fecIdx,int sck,int transport,BYTE *data);
unsigned short execRequest(CLN *cln);
int doScheduler(struct timeval *thisTime);
int doDelivery(struct timeval *thisTIme);
int sendToPeer(CLN *c,BYTE *payload,int msgtype);
int doRequest(int num, int *i);
int getConsumerRequest(CLN *cln,BYTE *buf,int packetsize);
int getConTblEntry(DTYPE *dout,DTYPE *din,char *devcontxt,char *devgrp,char *devname,char *property);
void displayData(int format, int fec, int dir);
int dumpConsumerList(void);
int dumpContractList(void);
int dumpEqmList(void);
int dumpEqmProperties(char *eqm);
int dumpEqmDevices(char *eqm);
int dumpConnectionList(void);
void dumpRegisteredUsers(void);
int dumpSemaphores(void);
void consumeData(BYTE *buf);
int addFecToList(char   *name);
int initFecTable(void);
int datacmp(BYTE *d1,BYTE *d2,int siz,int fmt,double t);
int initSendNetGlobalsList(void);
void getGlobalUpdate(BYTE *buf);
void quitServer(void);
void hardDelay(int msec);
int IsMemberLockSet(CLN *cln);
CLN *getNetworkClientFromList(void);
CLN *LocateConsumerInList(struct sockaddr_in *ip,IPXAddress *nwAddr,BYTE *immAddr);
TINE_EXPORT UINT32 ElfHash(const unsigned char *key);
void makeEqpDevicesHashTable(ExportListStruct *el);
int makeStockPropXRefTable(void);
#define GATE_BOOLEAN 0x01
#define GATE_FILTER 0x02
short IsGatedModifier(CONTRACT *con,char *baseProperty,UINT32 *gate);
short IsStockProperty(char *Property);
int lookupPropertyName(ExportListStruct *el,char *prpName,char *devName);
char *lookupPropertyId(ExportListStruct *el,short id);
int sndRegisteredExports(struct timeval *tv);
void ParsePropertyDescription(char *str,ExportPropertyListStruct *prp);
void ParseRedirector(char *str,PrpRedirBlk *prd);
ExportListStruct *getExportListItem(char *eqm);
ExportListStruct *getExportListItemFromId(int id);
int SystemStartGlobalSynchronization(void);
TINE_EXPORT int reformat(BYTE *dstData,int dstFmt,BYTE *srcData,int srcFmt,int num);
int getSrvAddress(DTYPE *sa,char *eqm,char *devsrv);
TINE_EXPORT int GetServerInitialization(void);

int IsInForeground(void);
void checkIPXNets(struct timeval *tv);
int GetSPXStream(WORD ID,CLN *cln);
int initIPX(void);
void CleanupIPX(struct timeval *tv);
void closeIPXServer(void);
void AcceptIPX(struct timeval *tv);
void AcceptSPXStream(void);
void CloseSPXStream(WORD ID);
int sendIPXData(CLN *cln, BYTE *data);
int sendIPXgcast(BYTE *buf,int size);
void fulfillSPXStreamRequest(WORD ID);
CLN *getConsumerIPXAddress(void);
char *ipxnet_ntoa(BYTE *node);
void dumpIPXNets(void);

void NotifyClient(int i,int cc);
void NotifyGlobalClient(int i,int cc);
int doTCPClient(int fec);
int listenUDP(int *s,unsigned short port,int initMultiCast);
int sendUDPBuffer(struct sockaddr_in *client, int msgsize);
int connectTCP(int port,int i);
int initTCP(void);
int initClient(void);
int GetTCPStream(int socket,CLN *cln);
void CloseTCPStream(int socket);
int sendIPData(CLN *cln,BYTE *data);
int sendIPgcast(BYTE *buf,int size);
void fulfillTCPStreamRequest(int socket);
CLN *getConsumerIPAddress(int sck,int prot);
int redirectLink(char *con,char *srv,char *prp,char *dev);
void AcceptTCPStream(void);
void dumpIPNets(void);
void AcceptIP(struct timeval *tv);
void CleanupIP(struct timeval *tv);
void closeIPServer(void);
int IsMemberIPNets(int num,struct sockaddr_in *lst,struct sockaddr_in *ipaddr);
int tUDPCall(int fecindex,BYTE *buf,int timout);
int tIPXCall(int fecindex,BYTE *buf,int timout);
int tExecLink(char *fec,char *eqm,char *dev,char *prp,DTYPE *dout,DTYPE *din,short access,short timout);
void checkClientNotification(void);
int readSelfTest(char *eqm,char *textbuf,int sizeInBytes);
int readFecLog(char *textbuf,int sizeInBytes);
int readFecLogFile(char *fname,char *textbuf,int sizeInBytes);
int writeFecLogFile(char *fname,char *textbuf,int sizeInBytes);
int readFecFile(char *fname,char *buf,int *sizeInBytes);
int writeFecFile(char *fname,char *buf,int sizeInBytes);
int scanFecLogFiles(char *path,char *subpath,NAME64 *namebuf,int *numfiles,int numoffset);
void applyFecLogPath(char *path,int pathlen);
void checkAccessLocks(time_t t);
int isDeniedUser(char *usr);
TINE_EXPORT void haltServer(void);
void freeServerMemory(void);
TINE_EXPORT void FreeEQPModules(void);
TINE_EXPORT int dbglog(char *text,...);
int _datcmp(char *d1,char *d2,int dsz,short fmt,DataToleranceFactor *dtf);
int isSpectrumArray(char *eqm,char *prpName);
int isChannelArray(char *eqm,char *prpName);
void getWildcardTokens(char *wildcardString,char *beginToken,char *endToken,char *midToken);
int wildcardMatches(char *devName,char *beginToken,int beginLen,char *endToken,int endLen,char *midToken,int midLen);
int isContiguous(char *eqm,char *devName,NAME64 *devList,int *start,int *len);
int isLocal(char *eqm,char *devName,char *prpName);
ExportPropertyListStruct *GetProperyListStruct(char *eqm,char *prpName);
int queryBitfieldFormat(char *srv,char *tag,NAME16I *bffq,int *bffqsiz);
int callEqmInSeparateThread(char *eqm,char *prp);
extern int gClnRecvQueueDepth;
int callEqm(ContractListStruct *cl,char *devName,char *devProperty,DTYPE *dout,DTYPE *din,short access);

TINE_EXPORT void ResetLastMinuteInitialization(void);
TINE_EXPORT void resetLastMinuteInitialization(void);

/* getLastContractId() */
/**
 * \brief Returns the last contract identifier.
 * 
 * A server can retieve the last contract identifier at any time.
 * 
 * \note A contract table identifier is a dynamic thing, to the extent that if a contract
 * comes and goes (is not persistent) it can have a different ID when it comes again.
 *
 * \return a valid contract ID or -1 if the contract table is empty.
 *
 */
TINE_EXPORT int GetLastContractId(void);

/* FindServerOnNetwork() */
/**
 * \brief Issues a multicast (or broadcast) to which configured TINE central servers respond
 * 
 * Diskless (or otherwise 'in-a-box') servers which need to find central servers such as the
 * Equipment Name Server, Central Alarm Server, or Post-Mortem Server, etc. can issue this
 * call to ascertain the address.
 *  
 * \param context (optional) the Context of the desired central server
 * \param eqmName the local equipment module name of the desired central server (e.g. "ENSEQM")
 * \param exportName the exported device server name of the desired central server. Note either
 * exportName or eqmName, but not both are optional input parameters.
 * \param fec (optional) is the returned FEC address structure of the server responding to the
 * call
 * \param srv (optional) is the returned Device Server Data address structure of the server responding
 * to the call.
 *
 * \return 0 if successful, otherwise a TINE completion code
 *
 */
TINE_EXPORT int FindServerOnNetwork(char *context,char *eqmName,char *exportName,FecDataStruct *fec,ExpDataStruct *srv);

extern int gSystemDataStamp;

TINE_EXPORT FecDataStruct *GetFecDataFromCache(char *context,char *expName);
TINE_EXPORT int GetFECAddr(char *fecName,FecDataStruct *fecaddr);
TINE_EXPORT int GetServerAddr(char *context,char *expName,ExpDataStruct *srvaddr,FecDataStruct *fecaddr);
TINE_EXPORT int GetConnectionAddr(char *context,char *expName,char *devName,char *prpName);
TINE_EXPORT void CloseClients(int closesockets);
#define closeClients(a1) CloseClients(a1)
TINE_EXPORT void CloseServers(int closesockets);
#define closeServers(a1) CloseServers(a1)
TINE_EXPORT void UseMultiThreadedCycler(int value);
TINE_EXPORT void UseSynchronousNotification(int value);

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif  /* SRVCORE_H_ */
